#!BPY

""" Registration info for Blender menus:
Name: 'Ambrose3d (.amb)...'
Blender: 243
Group: 'Export'
Tip: 'Export to Ambrose3d simple binary format.'
"""
__author__ = "Jari Komppa"
__url__ = ("blender", "elysiun", "Author's site, http://iki.fi/sol")
__version__ = "0.5"

__bpydoc__ = """\
Simple exporter for Ambrose3D format.
"""

# ***** BEGIN GPL LICENSE BLOCK *****
#
# Script copyright (C) Jari Komppa
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------


import Blender
from Blender import *
import math, time, os, struct
import BPyMesh

# --------------------------------------------------------------------------

# Set this to '1' to output all sorts of debug data to the blender console
debug = 0

OPTION_EXPORT_ANIM = 1
OPTION_COLLAPSE_MODIFIERS = 1
OPTION_TRIANGULATE_STRATEGY = 0

# --------------------------------------------------------------------------

# Helper functions for file IO.
# Note that all data is written in little-endian format.

def write_uint(file, data):
  file.write(struct.pack("<L", data))

def write_ushort(file, data):
  file.write(struct.pack("<H", data))
  
def write_char(file, data):
  file.write(struct.pack("<c", data))

def write_byte(file, data):
  file.write(struct.pack("<B", data))
  
def write_float(file, data):
  file.write(struct.pack("<f", data))

def write_matrix(file, m):
  data = [m[0][0], m[0][1], m[0][2], m[0][3], m[1][0], m[1][1], m[1][2], m[1][3], m[2][0], m[2][1], m[2][2], m[2][3], m[3][0], m[3][1], m[3][2], m[3][3]]
  file.write(struct.pack("<ffffffffffffffff", *data))

# --------------------------------------------------------------------------

# Helper functions to calculate the id:s of materials and textures.
# These are then used in the 3d engine to create pointers.

def materialid(mat):
  allmaterials = Blender.Material.Get()
  id = 0  
  for m in allmaterials:
    id += 1
    if m == mat:
      return id
  return 0

def textureid(tex):
  alltextures = Blender.Texture.Get()
  id = 0  
  for t in alltextures:
    id += 1
    if t == tex:
      return id
  return 0
      

# --------------------------------------------------------------------------

def write_header(file):
  global OPTION_EXPORT_ANIM
  
  if debug == 1:
    print("header..")
  
  # Header format tag
  write_char(file, 'A')
  write_char(file, 'm')
  write_char(file, 'b')
  write_byte(file, 1)
  
  scene = Blender.Scene.GetCurrent()
  renderdata = scene.getRenderingContext()  
  
  # Find out mesh, lamp and camera count
  
  meshes = 0
  lamps = 0
  cameras = 0  
  
  for obj in scene.objects:
    t = type(obj.getData())
    if debug == 1:
      print(obj.getName() + " " + obj.getType() + " ", t)
    if (t == Types.NMeshType):
      meshes+=1
    elif (t == Types.LampType):
      lamps+=1
    elif (t == Types.CameraType):
      cameras+=1
  
  # and write them out.
  
  if debug == 1:
  	print("Meshes: %d Lamps: %d Cameras: %d" % (meshes, lamps, cameras))
  
  write_uint(file,meshes)
  write_uint(file,lamps)
  write_uint(file,cameras)
  
  # Currently writing out animation controllers for everything.
  if OPTION_EXPORT_ANIM:
    write_uint(file,meshes+lamps+cameras) # anim controllers
  else:
    write_uint(file, 0)
        
  write_uint(file,len(Blender.Material.Get())) # number of materials
  
  write_uint(file,len(Blender.Texture.Get()))  # number of textures
  
  write_uint(file,renderdata.endFrame()) # max frames
  
  if debug == 1:
    print("header done")

# --------------------------------------------------------------------------



# --------------------------------------------------------------------------

def write_meshes(file):
  # Since hardware doesn't exactly like having one vertex with several UV coordinates,
  # we need to duplicate vertices so that there's a vertex for each UV coordinate.
  # The vertex duplication is currently done by adding new vertices to the end, which
  # may cause some problems down the road.. some ati hardware (at least) doesn't
  # like to render indices which delta is bigger than 64k.
  global OPTION_TRIANGULATE_STRATEGY, OPTION_COLLAPSE_MODIFIERS, debug
  scene = Blender.Scene.GetCurrent()
  id = 0
  for obj in scene.objects:
    id += 1
    mesh = obj.getData()
    t = type(mesh)
    if (t == Types.NMeshType):
      write_uint(file, id)
      write_matrix(file, obj.matrix)
      file.write(obj.getName())
      if debug == 1:
        print "---"
        print "Mesh '" + obj.getName() + "'"
      write_byte(file,0) # asciiz object name
      
      # Note: obj.getMaterials() doesn't work (blender 243)
      materials = mesh.materials
      if len(materials) == 0:
        write_uint(file, 0)
        if debug == 1:
          print "no materials?!?!"
      else:
		    # Only using the first material - if user uses material number two somehow and
		    # leaves material one to null, this will cause problems.
        write_uint(file, materialid(materials[0]))
        if debug == 1:
          print "material '" + materials[0].name + "' id %d" % materialid(materials[0])
      
      # (possibly) collapse submeshes here:
      m = BPyMesh.getMeshFromObject(obj,None,OPTION_COLLAPSE_MODIFIERS,False,scene)
      verts = m.verts
      faces = m.faces      
      facecount = 0
  
  		# Calculate the number of triangles. Quads are split into two triangles.
      for f in faces:
        if len(f.v) == 3:
          facecount+=1
        elif len(f.v) == 4:
          facecount+=2

			# Temporary storage for vertex coordinate and vertex normal data.      
      vertcodata = []
      vertnodata = []

			# First copy the originals
      for i in xrange(len(verts)):
        vertcodata.append(verts[i].co)
        vertnodata.append(verts[i].no)     
      
      # Temporary storage for U and V coordinates (start off with 0.0)
      ulist = [0.0] * len(verts)
      vlist = [0.0] * len(verts)
      
      # Temporary index and face uv containers
      idxlist = []
      faceuv = []      
      
      # Grab face UVs and create indices based on quad splitting strategy
      for f in faces:
        if len(f.v) == 3:
          if m.faceUV:
            faceuv.append(f.uv[0])
            faceuv.append(f.uv[1])
            faceuv.append(f.uv[2])
          idxlist.append(f.v[0].index)
          idxlist.append(f.v[1].index)
          idxlist.append(f.v[2].index)
        elif len(f.v) == 4:
          shortest = (f.v[0].co - f.v[2].co).length < (f.v[1].co - f.v[3].co).length          

          if OPTION_TRIANGULATE_STRATEGY == 1:
            shortest = not shortest;

          if (OPTION_TRIANGULATE_STRATEGY == 0) or (OPTION_TRIANGULATE_STRATEGY == 1):
            if shortest:
              if m.faceUV:
                faceuv.append(f.uv[0])
                faceuv.append(f.uv[1])
                faceuv.append(f.uv[2])
                faceuv.append(f.uv[0])
                faceuv.append(f.uv[2])
                faceuv.append(f.uv[3])
              idxlist.append(f.v[0].index)
              idxlist.append(f.v[1].index)
              idxlist.append(f.v[2].index)
              idxlist.append(f.v[0].index)
              idxlist.append(f.v[2].index)
              idxlist.append(f.v[3].index)
            else:
              if m.faceUV:
                faceuv.append(f.uv[0])
                faceuv.append(f.uv[1])
                faceuv.append(f.uv[3])
                faceuv.append(f.uv[1])
                faceuv.append(f.uv[2])
                faceuv.append(f.uv[3])
              idxlist.append(f.v[0].index)
              idxlist.append(f.v[1].index)
              idxlist.append(f.v[3].index)
              idxlist.append(f.v[1].index)
              idxlist.append(f.v[2].index)
              idxlist.append(f.v[3].index)

          if (OPTION_TRIANGULATE_STRATEGY == 3):
            if m.faceUV:
              faceuv.append(f.uv[0])
              faceuv.append(f.uv[1])
              faceuv.append(f.uv[2])
              faceuv.append(f.uv[0])
              faceuv.append(f.uv[2])
              faceuv.append(f.uv[3])
            idxlist.append(f.v[0].index)
            idxlist.append(f.v[1].index)
            idxlist.append(f.v[2].index)
            idxlist.append(f.v[0].index)
            idxlist.append(f.v[2].index)
            idxlist.append(f.v[3].index)

          if (OPTION_TRIANGULATE_STRATEGY == 4):
            if m.faceUV:
              faceuv.append(f.uv[0])
              faceuv.append(f.uv[1])
              faceuv.append(f.uv[3])
              faceuv.append(f.uv[1])
              faceuv.append(f.uv[2])
              faceuv.append(f.uv[3])
            idxlist.append(f.v[0].index)
            idxlist.append(f.v[1].index)
            idxlist.append(f.v[3].index)
            idxlist.append(f.v[1].index)
            idxlist.append(f.v[2].index)
            idxlist.append(f.v[3].index)
      
      # Next handle UV coordinates.
      dupedvtxs = 0
      if m.vertexUV:
      	# In case of vertex "sticky" UVs, things are simple: one uv pair per vertex.
        if debug == 1:
          print "object has vertex UVs"
        for i in xrange(len(verts)):
          ulist[i] = verts[i].uvco[0]
          vlist[i] = verts[i].uvco[1]
      elif m.faceUV:
      	# Face UVs (the most typical) are more complicated, as one vertex most likely has several UV
      	# coordinates.
        if debug == 1:
          print "object has face UVs - possibly need to duplicate vertices"
        # A list of flags to check if this vertex already has UVs
        gotuvlist = [False] * len(verts)

        for i in xrange(len(idxlist)):          
          if gotuvlist[idxlist[i]] and not (ulist[idxlist[i]] == faceuv[i][0] and vlist[idxlist[i]] == faceuv[i][1]):

            # need to duplicate
            vertcodata.append(vertcodata[idxlist[i]])
            vertnodata.append(vertnodata[idxlist[i]])
            
            # and append the uv:s in the list.
            ulist.append(faceuv[i][0])
            vlist.append(faceuv[i][1])
            # one inserted index..
            dupedvtxs += 1
            # move index to a new one
            idxlist[i] = len(vertcodata) - 1
            
            # also need to append to the helper lists
            gotuvlist.append(True)
            
          else:
          	# No need to duplicate (yet) - record the UV and flag vertex as done
            ulist[idxlist[i]] = faceuv[i][0]
            vlist[idxlist[i]] = faceuv[i][1]
            gotuvlist[idxlist[i]] = True
      else:
        # no UV coordinates - generate some.
        if debug == 1:
          print "object has no UVs, generating from normal"
        for i in xrange(len(verts)):
          ulist[i] = verts[i].no[0] + verts[i].no[2]
          ulist[i] = verts[i].no[1] + verts[i].no[2]
        
      if debug == 1:
        print "Vertices: %d realverts:%d Faces:%d Realfaces:%d uv-duped verts:%d" % (len(verts), len(verts)+dupedvtxs, len(faces), facecount, dupedvtxs)

      write_uint(file, len(verts) + dupedvtxs)
      write_uint(file, facecount)      

      # outvertcount is only used for debugging      
      outvertcount = 0
      for k in vertcodata:
        write_float(file, k[0])
        write_float(file, k[1])
        write_float(file, k[2])
        outvertcount += 1
      if debug == 1:
        print "%d vertices written" % outvertcount
      outvertcount = 0
      for k in vertnodata:
        write_float(file, k[0])
        write_float(file, k[1])
        write_float(file, k[2])
        outvertcount += 1
      if debug == 1:
        print "%d normals written" % outvertcount
      outvertcount = 0
      for i in xrange(len(ulist)):
        write_float(file, ulist[i])
        write_float(file, 1.0 - vlist[i])
        outvertcount += 1
      if debug == 1:
        print "%d UVs written" % outvertcount
      for i in idxlist:
        write_uint(file, i)
      
# --------------------------------------------------------------------------

def write_lamps(file):
	# Currently handles all lights as omnis.
  scene = Blender.Scene.GetCurrent()
  id = 0
  for obj in scene.objects:
    lamp = obj.getData()
    id += 1
    t = type(lamp)
    if (t == Types.LampType):
      write_uint(file, id)
      write_matrix(file, obj.matrix)
      file.write(obj.getName())
      write_byte(file, 0)
      write_float(file, lamp.col[0])
      write_float(file, lamp.col[1])
      write_float(file, lamp.col[2])
      
# --------------------------------------------------------------------------

def write_cameras(file):
  scene = Blender.Scene.GetCurrent()
  id = 0
  for obj in scene.objects:
    camera = obj.getData()
    id += 1
    t = type(camera)
    if (t == Types.CameraType):
      write_uint(file, id)
      m = obj.matrix.copy()
      m.resize4x4()
      m.invert()
      write_matrix(file, m)
      file.write(obj.getName())
      write_byte(file, 0)
      write_float(file, camera.clipEnd)
      write_float(file, camera.clipStart)
      write_float(file, camera.lens)

def write_anims(file):
	# Animations are stored as target object/camera/lamp id and one matrix
	# per frame.
  # Instead of writing all matrices for everything, probably
  # should detect non-moving objects, and only export changing
  # things for the moving ones.. and use pos/scale/rotation as 
  # vec3,vec3,quaternion..
  # The current way of storing animations takes way too much space.
  scene = Blender.Scene.GetCurrent()
  renderdata = scene.getRenderingContext()
  animlen = renderdata.endFrame()
  id = 0
  for obj in scene.objects:
    odata = obj.getData()
    id += 1
    t = type(odata)
    if t == Types.CameraType or t == Types.LampType or t == Types.NMeshType:
      write_uint(file, id)
      for i in range(1,animlen+1):
        renderdata.currentFrame(i)
        if t == Types.CameraType:
          m = obj.matrix.copy()
          m.resize4x4()
          m.invert()
          write_matrix(file, m)
        else:
          write_matrix(file, obj.matrix)
          

# --------------------------------------------------------------------------

def write_textures(file):
	# For textures, the texture name and (for image textures) filename are stored.
	# Other types of textures could be easily supported.
  alltextures = Blender.Texture.Get()
  for tex in alltextures:
    file.write(tex.name)
    write_byte(file, 0) # asciiz texture name
    write_uint(file, tex.type)
    if tex.type == Blender.Texture.Types.IMAGE:
      file.write(tex.getImage().getFilename()[2:]) # For some reason blender prefixes the filename with // - the [2:] skips this
      write_byte(file, 0) # asciiz filename
    
# --------------------------------------------------------------------------
  
def write_mtex(file, mtex):
	# Several materials may use one texture and textures can be used
	# in many ways.. the mtex handles this binding.
  write_uint(file, mtex.blendmode)
  write_uint(file, mtex.mapto)
  write_uint(file, textureid(mtex.tex))
    
# --------------------------------------------------------------------------
  
def write_materials(file):
	# Only a few material properties are exported for now
  allmaterials = Blender.Material.Get()
  for mat in allmaterials:
    file.write(mat.name)
    write_byte(file, 0) # asciiz material name
    write_float(file, mat.alpha)
    write_float(file, mat.rgbCol[0])
    write_float(file, mat.rgbCol[1])
    write_float(file, mat.rgbCol[2])
    write_float(file, mat.specCol[0])
    write_float(file, mat.specCol[1])
    write_float(file, mat.specCol[2])
    
    texlinks = mat.getTextures()
    count = 0
    for t in texlinks:
      if t != None:
        count += 1
    write_uint(file, count)
    for t in texlinks:
      if t != None:
        write_mtex(file, t)
    

def do_export(filename):
	# Perform the actual exporting.
	
  global OPTION_EXPORT_ANIM
  
  # We're checking how long the export takes
  start = time.clock()
  
  try:
    file = open(filename, 'wb')
  except IOError, (errno, strerror):
    error = "IOError #%s: %s" % (errno, strerror)
    REPORT_DATA['errors'].append("Saving failed - %s." % error)
    ERROR_MSG = "Couldn't save file!%%t|%s" % error
    Draw.PupMenu("File open failed")
    Draw.Exit()
    return None

  scene = Blender.Scene.GetCurrent()
  renderdata = scene.getRenderingContext()

	# Take note of the current frame in the editor (storing the animations changes the frame)
  currframe = renderdata.currentFrame()
  write_header(file)
  write_textures(file)
  write_materials(file)
  write_meshes(file)
  write_lamps(file)
  write_cameras(file)
  if OPTION_EXPORT_ANIM:
    write_anims(file)
  renderdata.currentFrame(currframe) # return to stored frame
  file.close()
  
  # Report the passed time in a popup window
  end = time.clock()
  seconds = " in %.2f %s" % (end-start, "seconds")
  message = "Successfully exported " + os.path.basename(filename) + seconds
  Draw.PupMenu("Ambrose3d export done|" + message + "%x0")
  Draw.Exit()

# --------------------------------------------------------------------------

def fs_callback(filename):
  if filename.find('.amb', -4) <= 0: filename += '.amb'
  do_export(filename)

# --------------------------------------------------------------------------

def event(evt, val):    # the function to handle input events

  if evt == Draw.ESCKEY:
    Draw.Exit()                 # exit when user presses ESC
    return

  Draw.Register(gui, event, button_event)  # re-register to stay in the loop

# --------------------------------------------------------------------------

def button_event(evt):  # the function to handle Draw Button events
  global OPTION_EXPORT_ANIM, OPTION_COLLAPSE_MODIFIERS, OPTION_TRIANGULATE_STRATEGY
  if evt == 1:
    Blender.Window.FileSelector(fs_callback, "Export AMB", Blender.sys.makename(ext = ".amb"))    
  elif evt == 2:
    OPTION_EXPORT_ANIM = 1 - OPTION_EXPORT_ANIM
    Draw.Redraw(1)
  elif evt == 3:
    OPTION_COLLAPSE_MODIFIERS = 1 - OPTION_COLLAPSE_MODIFIERS
    Draw.Redraw(1)
  elif evt == 4:
    OPTION_TRIANGULATE_STRATEGY = 0;
    Draw.Redraw(1)
  elif evt == 5:
    OPTION_TRIANGULATE_STRATEGY = 1;
    Draw.Redraw(1)
  elif evt == 6:
    OPTION_TRIANGULATE_STRATEGY = 2;
    Draw.Redraw(1)
  elif evt == 7:
    OPTION_TRIANGULATE_STRATEGY = 3;
    Draw.Redraw(1)
  else:
    Draw.Register(gui, event, button_event)

# --------------------------------------------------------------------------

def gui(): 
  global OPTION_EXPORT_ANIM, OPTION_COLLAPSE_MODIFIERS, OPTION_TRIANGULATE_STRATEGY
  BGL.glClearColor(0.5,0.5,0.5,1)
  BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
  BGL.glColor3f(1,1,1)
  BGL.glRasterPos2i(10, 200)
  Draw.Text("Abrose3d exporter " + __version__, "large")
  BGL.glRasterPos2i(10, 180)
  Draw.Text("http://iki.fi/sol", "large")
  
  Draw.Toggle("Export Anims", 2, 10, 100, 155, 20, OPTION_EXPORT_ANIM,"Export animations")
  BGL.glRasterPos2i(172, 106)
  if OPTION_EXPORT_ANIM: 
    Draw.Text("Animations will be exported", "small")
  else: 
    Draw.Text("Animations will not be exported", "small")

  Draw.Toggle("Collapse", 3, 10, 70, 155, 20, OPTION_COLLAPSE_MODIFIERS,"Collapse submeshes")
  BGL.glRasterPos2i(172, 76)
  if OPTION_COLLAPSE_MODIFIERS: 
    Draw.Text("Submeshes will be collapsed", "small")
  else: 
    Draw.Text("Submeshes will not be collapsed", "small")


  BGL.glRasterPos2i(400, 170)
  Draw.Text("Quad to triangle conversion strategy", "large")
  Draw.Toggle("Short edge", 4, 400, 130, 155, 20, OPTION_TRIANGULATE_STRATEGY == 0,"Pick shorter edge")
  Draw.Toggle("Long edge", 5, 400, 110, 155, 20, OPTION_TRIANGULATE_STRATEGY == 1,"Pick longer edge")
  Draw.Toggle("Pick cw", 6, 400, 90, 155, 20, OPTION_TRIANGULATE_STRATEGY == 2,"Pick clockwise always")
  Draw.Toggle("Pick ccw", 7, 400, 70, 155, 20, OPTION_TRIANGULATE_STRATEGY == 3,"Pick counterclockwise always")


  Draw.Button("> Export <", 1, 10, 16, 120, 20, "Perform export")

# --------------------------------------------------------------------------

Draw.Register(gui, event, button_event)  # registering the 3 callbacks
