// Simple example of Ambrose3d use

#include "Ambrose.h"

// Scene to be played
AmbroseScene scn;

// 'Agent' image texture handle
GLuint agentid;

// Handle SDL events
static void process_events()
{
    SDL_Event event;

    while (SDL_PollEvent(&event)) 
    {
        switch (event.type) 
        {
        case SDL_KEYDOWN:
            switch (event.key.keysym.sym) 
            {
            case SDLK_ESCAPE:
                SDL_Quit();
                exit(0);
                break;
            default:
                break;
            }
            break;
        case SDL_QUIT:
            SDL_Quit();
            exit(0);
            break;
        }
    }
}

// Draw the 'agent' image in the top right corner
void draw_agent()
{
    // Push GL_PROJECTION matrix, and set it to orthographic mode for 2d rendering
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(0,640,480,0,0,100);

    // Push GL_MODELVIEW matrix and set it to identity
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    // Push attributes
    glPushAttrib(~0);
    glPushClientAttrib(~0);

    // White
    glColor4f(1,1,1,1);
    // No z-buffering
    glDisable(GL_DEPTH_TEST);
    // Texture please
    glEnable(GL_TEXTURE_2D);
    // No lighting
    glDisable(GL_LIGHTING);
    // Enable blending
    glEnable(GL_BLEND);
    // Source alpha-based weighted average blending
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    // Use the agent texture
    glBindTexture(GL_TEXTURE_2D, agentid);
    // Since we're just drawing one quad, no sense in setting up vertex buffer
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0,0);
    glVertex2f(640-256,0);
    glTexCoord2f(1,0);
    glVertex2f(640,0);
    glTexCoord2f(0,1);
    glVertex2f(640-256,64);
    glTexCoord2f(1,1);
    glVertex2f(640,64);
    glEnd();

    // z-buffering back on
    glEnable(GL_DEPTH_TEST);  
    // and blending off
    glDisable(GL_BLEND);

    // pop attributes
    glPopClientAttrib();
    glPopAttrib();
    // pop modelview matrix
    glPopMatrix();
    // pop projection matrix
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
}

// Performs timing calculation on a N frame average, displays on window title bar
static void estimate_fps()
{
#define FRAMEAVERAGE 100
    static int times[FRAMEAVERAGE];
    static int latest = 0;  
    times[latest] = SDL_GetTicks();
    int oldest = latest + 1;
    if (oldest >= FRAMEAVERAGE) oldest = 0;
    char temp[200];
    if (times[oldest] != 0)
    {
        float mspf = (times[latest] - times[oldest]) / (float)FRAMEAVERAGE;
        sprintf(temp,"Ambrose3d %3.3fmspf %3.3ffps", mspf, 1000.0f / mspf); 
        SDL_WM_SetCaption(temp, NULL);
    }
    latest++;
    if (latest >= FRAMEAVERAGE)
        latest = 0;
}

// Do the rendering
static void draw_screen()
{
    float animation_pos = SDL_GetTicks() / 40000.0f;

    // Apply animations to the position based on tick
    scn.applyAnimations(animation_pos);
    // Set up lights (lights may have moved due to animation)
    scn.setupLights();
    // Set up camera 0 at aspect ratio 4/3
    scn.setupCamera(0, 4.0f / 3.0f);

    // Clear background
    glClearColor(0.0f,0.0f,0.2f,1.0f);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // Render the scene
    scn.render();

    // draw the 'agent' image in top right corner
    draw_agent();

    // display the result
    SDL_GL_SwapBuffers();

    // don't hog all cpu time..
    SDL_Delay(1);

    // and calculate the current FPS
    estimate_fps();
}

// Set up some OpenGL defaults
static void setup_opengl( int width, int height )
{
    // Smooth shading please (as opposed to flat)
    glShadeModel(GL_SMOOTH);

    // Cull back faces
    glCullFace(GL_BACK);
    // From face is the clockwise polygon
    glFrontFace(GL_CW);
    // Enable culling
    glEnable(GL_CULL_FACE);

    // Disable culling - the test scene has 2-sided polygons, but the
    // exported format does not currently support this
    glDisable(GL_CULL_FACE);

    // Enable z-buffering
    glEnable(GL_DEPTH_TEST);
    // with 'less or equal' test
    glDepthFunc(GL_LEQUAL);

    // Set up full screen viewport
    glViewport(0, 0, width, height);

}


int main (int argc, char** args)
{
    const SDL_VideoInfo *info = NULL;
    int width = 0;
    int height = 0;
    int bpp = 0;
    int flags = 0;

    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) 
    {
        fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    // Ask for current video mode info
    info = SDL_GetVideoInfo();

    if (!info) 
    {
        fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    // Could use width and height from the current mode, but we'll use 640x480 for now
    width = 640;
    height = 480;
    // Same bpp as current mode
    bpp = info->vfmt->BitsPerPixel;

    // Set up OpenGL attributes - 32bit RGBA color, 24bit depth buffer, double buffering
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    flags = SDL_OPENGL; //| SDL_FULLSCREEN;

    // Set up video
    if (SDL_SetVideoMode(width, height, bpp, flags) == 0) 
    {
        fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    // Set up some opengl states
    setup_opengl(width, height);

    // Load the scene
    if (!scn.load("boat11.amb"))
        exit(0);

    // Load the agent image
    agentid = loadTexture("ambrose3d_agent.png");

    // Main loop
    while (1) 
    {
        process_events();
        draw_screen();
    }

    return 0;
}
